@@ -632,6 +632,137 @@ impl<T: Deref<Target = str>> Iri<T> {
         self.0.resolve_into_unchecked(iri, target_buffer)
     }
 
+    /// Returns an IRI that, when resolved against the current IRI returns `abs`.
+    ///
+    /// This function returns an error
+    /// if is not possible to build a relative IRI that can resolve to the same IRI.
+    /// For example, when the path contains `/../`.
+    ///
+    /// Note that the output of this function might change in minor releases.
+    ///
+    /// ```
+    /// use oxiri::Iri;
+    ///
+    /// let base_iri = Iri::parse("http://foo.com/bar/baz")?;
+    /// let iri = Iri::parse("http://foo.com/bar/bat#foo")?;
+    /// let relative_iri = base_iri.relativize(&iri)?;
+    /// assert_eq!(relative_iri, "bat#foo");
+    /// # Result::<(), Box<dyn std::error::Error>>::Ok(())
+    /// ```
+    pub fn relativize<T2: Deref<Target = str>>(
+        &self,
+        abs: &Iri<T2>,
+    ) -> Result<IriRef<String>, IriRelativizeError> {
+        let base = self;
+        let abs_authority = abs.authority();
+        let base_authority = base.authority();
+        let abs_path = abs.path();
+        let base_path = base.path();
+        let abs_query = abs.query();
+        let base_query = base.query();
+
+        // We validate the path, resolving algorithm eats /. and /.. in hierarchical path
+        for segment in abs_path.split('/').skip(1) {
+            if matches!(segment, "." | "..") {
+                return Err(IriRelativizeError {});
+            }
+        }
+
+        if abs.scheme() != base.scheme()
+            || abs_authority.is_none() && base_authority.is_some()
+            || abs_path
+                // Might confuse with a scheme
+                .split_once(':')
+                .map_or(false, |(candidate_scheme, _)| {
+                    !candidate_scheme.contains('/')
+                })
+        {
+            return Ok(IriRef {
+                iri: abs.0.to_string(),
+                positions: abs.0.positions,
+            });
+        }
+        if abs_authority != base_authority
+            // the resolution algorithm does not handle empty paths:
+            || abs_path.is_empty() && (!base_path.is_empty() || base_query.is_some())
+            // confusion with authority:
+            || abs_path.starts_with("//")
+        {
+            return Ok(IriRef {
+                iri: abs.0[abs.0.positions.scheme_end..].to_string(),
+                positions: IriElementsPositions {
+                    scheme_end: 0,
+                    authority_end: abs.0.positions.authority_end - abs.0.positions.scheme_end,
+                    path_end: abs.0.positions.path_end - abs.0.positions.scheme_end,
+                    query_end: abs.0.positions.query_end - abs.0.positions.scheme_end,
+                },
+            });
+        }
+        if abs_path != base_path || abs_query.is_none() && base_query.is_some() {
+            let number_of_shared_characters = abs_path
+                .bytes()
+                .zip(base_path.bytes())
+                .take_while(|(l, r)| l == r)
+                .count();
+            // We decrease until finding a /
+            let number_of_shared_characters = abs_path[..number_of_shared_characters]
+                .rfind('/')
+                .map_or(0, |n| n + 1);
+            return if abs_path[number_of_shared_characters..].contains('/')
+                || base_path[number_of_shared_characters..].contains('/')
+                || abs_path[number_of_shared_characters..].is_empty()
+                || abs_path[number_of_shared_characters..].contains(':')
+            {
+                // We output the full path because we have a / or an empty end
+                Ok(IriRef {
+                    iri: abs.0[abs.0.positions.authority_end..].to_string(),
+                    positions: IriElementsPositions {
+                        scheme_end: 0,
+                        authority_end: 0,
+                        path_end: abs.0.positions.path_end - abs.0.positions.authority_end,
+                        query_end: abs.0.positions.query_end - abs.0.positions.authority_end,
+                    },
+                })
+            } else {
+                // We just override the last element
+                Ok(IriRef {
+                    iri: abs.0[abs.0.positions.authority_end + number_of_shared_characters..]
+                        .to_string(),
+                    positions: IriElementsPositions {
+                        scheme_end: 0,
+                        authority_end: 0,
+                        path_end: abs.0.positions.path_end
+                            - abs.0.positions.authority_end
+                            - number_of_shared_characters,
+                        query_end: abs.0.positions.query_end
+                            - abs.0.positions.authority_end
+                            - number_of_shared_characters,
+                    },
+                })
+            };
+        }
+        if abs_query != base_query {
+            return Ok(IriRef {
+                iri: abs.0[abs.0.positions.path_end..].to_string(),
+                positions: IriElementsPositions {
+                    scheme_end: 0,
+                    authority_end: 0,
+                    path_end: 0,
+                    query_end: abs.0.positions.query_end - abs.0.positions.path_end,
+                },
+            });
+        }
+        Ok(IriRef {
+            iri: abs.0[abs.0.positions.query_end..].to_string(),
+            positions: IriElementsPositions {
+                scheme_end: 0,
+                authority_end: 0,
+                path_end: 0,
+                query_end: 0,
+            },
+        })
+    }
+
     /// Returns an IRI borrowing this IRI's text
     #[inline]
     pub fn as_ref(&self) -> Iri<&str> {
@@ -993,7 +1124,7 @@ impl<'de, T: Deref<Target = str> + Deserialize<'de>> Deserialize<'de> for Iri<T>
     }
 }
 
-/// An error raised during [`Iri`](struct.Iri.html) validation.
+/// An error raised during [`Iri`] or [`IriRef`] validation.
 #[derive(Debug)]
 pub struct IriParseError {
     kind: IriParseErrorKind,
@@ -1040,6 +1171,24 @@ enum IriParseErrorKind {
     InvalidPercentEncoding([Option<char>; 3]),
 }
 
+/// An error raised when calling [`Iri::relativize`].
+///
+/// It can happen when it is not possible to build a relative IRI that can resolve to the same IRI.
+/// For example, when the path contains `/../`.
+#[derive(Debug)]
+pub struct IriRelativizeError {}
+
+impl fmt::Display for IriRelativizeError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(
+            f,
+            "It is not possible to make this IRI relative because it contains `/..` or `/.`"
+        )
+    }
+}
+
+impl Error for IriRelativizeError {}
+
 #[derive(Debug, Clone, Copy)]
 struct IriElementsPositions {
     scheme_end: usize,
@@ -1297,7 +1446,7 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                         // We have some path or authority, we keep a base '/'
                         self.output.push('/');
                     }
-                    self.parse_path()
+                    self.parse_relative_path()
                 }
             }
         } else {
@@ -1309,11 +1458,22 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                 self.parse_path_or_authority()
             } else {
                 self.output_positions.authority_end = 0;
-                self.parse_path()
+                self.parse_relative_path()
             }
         }
     }
 
+    fn parse_relative_path(&mut self) -> Result<(), IriParseError> {
+        while let Some(c) = self.input.front() {
+            if matches!(c, '/' | '?' | '#') {
+                break;
+            }
+            self.input.next();
+            self.read_url_codepoint_or_echar(c, |c| is_iunreserved_or_sub_delims(c) || c == '@')?;
+        }
+        self.parse_path()
+    }
+
     fn parse_relative_slash(&mut self, base: &IriRef<&'a str>) -> Result<(), IriParseError> {
         if self.input.starts_with('/') {
             self.input.next();
@@ -1350,7 +1510,9 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                     return self.parse_host();
                 }
                 Some(c) => {
-                    self.read_url_codepoint_or_echar(c)?;
+                    self.read_url_codepoint_or_echar(c, |c| {
+                        is_iunreserved_or_sub_delims(c) || c == ':'
+                    })?;
                 }
             }
         }
@@ -1365,7 +1527,9 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                 if c == ']' {
                     let ip = &self.iri[start_position + 1..self.input.position - 1];
                     if !UNCHECKED {
-                        if let Err(error) = Ipv6Addr::from_str(ip) {
+                        if ip.starts_with('v') || ip.starts_with('V') {
+                            self.validate_ip_v_future(ip)?;
+                        } else if let Err(error) = Ipv6Addr::from_str(ip) {
                             return self.parse_error(IriParseErrorKind::InvalidHostIp(error));
                         }
                     }
@@ -1411,7 +1575,7 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                         self.output_positions.authority_end = self.output.len();
                         return self.parse_path_start(c);
                     }
-                    Some(c) => self.read_url_codepoint_or_echar(c)?,
+                    Some(c) => self.read_url_codepoint_or_echar(c, is_iunreserved_or_sub_delims)?,
                 }
             }
         }
@@ -1459,7 +1623,9 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                 self.parse_path()
             }
             Some(c) => {
-                self.read_url_codepoint_or_echar(c)?;
+                self.read_url_codepoint_or_echar(c, |c| {
+                    is_iunreserved_or_sub_delims(c) || matches!(c, ':' | '@')
+                })?;
                 self.parse_path()
             }
         }
@@ -1496,7 +1662,9 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                         return Ok(());
                     }
                 }
-                Some(c) => self.read_url_codepoint_or_echar(c)?,
+                Some(c) => self.read_url_codepoint_or_echar(c, |c| {
+                    is_iunreserved_or_sub_delims(c) || matches!(c, ':' | '@')
+                })?,
             }
         }
     }
@@ -1508,7 +1676,9 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
                 self.output.push('#');
                 return self.parse_fragment();
             } else {
-                self.read_url_query_codepoint_or_echar(c)?
+                self.read_url_codepoint_or_echar(c, |c| {
+                    is_iunreserved_or_sub_delims(c) || matches!(c, ':' | '@' | '/' | '?' | '\u{E000}'..='\u{F8FF}' | '\u{F0000}'..='\u{FFFFD}' | '\u{100000}'..='\u{10FFFD}')
+                })?
             }
         }
         self.output_positions.query_end = self.output.len();
@@ -1517,7 +1687,9 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
 
     fn parse_fragment(&mut self) -> Result<(), IriParseError> {
         while let Some(c) = self.input.next() {
-            self.read_url_codepoint_or_echar(c)?
+            self.read_url_codepoint_or_echar(c, |c| {
+                is_iunreserved_or_sub_delims(c) || matches!(c, ':' | '@' | '/' | '?')
+            })?;
         }
         Ok(())
     }
@@ -1530,9 +1702,12 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
             .truncate(last_slash_position + self.output_positions.authority_end)
     }
 
-    #[inline]
-    fn read_url_codepoint_or_echar(&mut self, c: char) -> Result<(), IriParseError> {
-        if UNCHECKED || is_url_code_point(c) {
+    fn read_url_codepoint_or_echar(
+        &mut self,
+        c: char,
+        valid: impl Fn(char) -> bool,
+    ) -> Result<(), IriParseError> {
+        if UNCHECKED || valid(c) {
             self.output.push(c);
             Ok(())
         } else if c == '%' {
@@ -1542,19 +1717,6 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
         }
     }
 
-    #[inline]
-    fn read_url_query_codepoint_or_echar(&mut self, c: char) -> Result<(), IriParseError> {
-        if UNCHECKED || is_url_query_code_point(c) {
-            self.output.push(c);
-            Ok(())
-        } else if c == '%' {
-            self.read_echar()
-        } else {
-            self.parse_error(IriParseErrorKind::InvalidIriCodePoint(c))
-        }
-    }
-
-    #[inline]
     fn read_echar(&mut self) -> Result<(), IriParseError> {
         let c1 = self.input.next();
         let c2 = self.input.next();
@@ -1574,14 +1736,51 @@ impl<'a, O: OutputBuffer, const UNCHECKED: bool> IriParser<'a, O, UNCHECKED> {
         }
     }
 
-    #[inline]
     fn parse_error<T>(&self, kind: IriParseErrorKind) -> Result<T, IriParseError> {
         Err(IriParseError { kind })
     }
+
+    // IPvFuture      = "v" 1*HEXDIG "." 1*( unreserved / sub-delims / ":" )
+    fn validate_ip_v_future(&self, ip: &str) -> Result<(), IriParseError> {
+        let mut chars = ip.chars();
+
+        let c = chars.next().ok_or(IriParseError {
+            kind: IriParseErrorKind::InvalidHostCharacter(']'),
+        })?;
+        if !matches!(c, 'v' | 'V') {
+            return self.parse_error(IriParseErrorKind::InvalidHostCharacter(c));
+        };
+
+        let mut with_a_version = false;
+        for c in &mut chars {
+            if c == '.' {
+                break;
+            } else if c.is_ascii_hexdigit() {
+                with_a_version = true;
+            } else {
+                return self.parse_error(IriParseErrorKind::InvalidHostCharacter(c));
+            }
+        }
+        if !with_a_version {
+            return self.parse_error(IriParseErrorKind::InvalidHostCharacter(
+                chars.next().unwrap_or(']'),
+            ));
+        }
+
+        if chars.as_str().is_empty() {
+            return self.parse_error(IriParseErrorKind::InvalidHostCharacter(']'));
+        };
+        for c in chars {
+            if !is_unreserved_or_sub_delims(c) && c != ':' {
+                return self.parse_error(IriParseErrorKind::InvalidHostCharacter(c));
+            }
+        }
+
+        Ok(())
+    }
 }
 
-#[inline]
-fn is_url_code_point(c: char) -> bool {
+fn is_iunreserved_or_sub_delims(c: char) -> bool {
     matches!(c,
         'a'..='z'
         | 'A'..='Z'
@@ -1597,12 +1796,8 @@ fn is_url_code_point(c: char) -> bool {
         | ','
         | '-'
         | '.'
-        | '/'
-        | ':'
         | ';'
         | '='
-        | '?'
-        | '@'
         | '_'
         | '~'
         | '\u{A0}'..='\u{D7FF}'
@@ -1625,8 +1820,25 @@ fn is_url_code_point(c: char) -> bool {
     )
 }
 
-#[inline]
-fn is_url_query_code_point(c: char) -> bool {
-    is_url_code_point(c)
-        || matches!(c, '\u{E000}'..='\u{F8FF}' | '\u{F0000}'..='\u{FFFFD}' | '\u{100000}'..='\u{10FFFD}')
+fn is_unreserved_or_sub_delims(c: char) -> bool {
+    matches!(c,
+        'a'..='z'
+        | 'A'..='Z'
+        | '0'..='9'
+        | '!'
+        | '$'
+        | '&'
+        | '\''
+        | '('
+        | ')'
+        | '*'
+        | '+'
+        | ','
+        | '-'
+        | '.'
+        | ';'
+        | '='
+        | '_'
+        | '~'
+    )
 }
