@@ -33,13 +33,24 @@ fn test_parsing() {
         "http://example.com/foo/bar/?q=1&r=2#toto",
         "http://example.com/foo/bar/.././baz",
         "http://a.example/AZaz\u{00C0}\u{00D6}\u{00D8}\u{00F6}\u{00F8}\u{02FF}\u{0370}\u{037D}\u{037F}\u{1FFF}\u{200C}\u{200D}\u{2070}\u{218F}\u{2C00}\u{2FEF}\u{3001}\u{D7FF}\u{FA0E}\u{FDCF}\u{FDF0}\u{FFEF}\u{10000}\u{EFFFD}",
-        "http://a.example/?AZaz\u{E000}\u{F8FF}\u{F0000}\u{FFFFD}\u{100000}\u{10FFFD}\u{00C0}\u{00D6}\u{00D8}\u{00F6}\u{00F8}\u{02FF}\u{0370}\u{037D}\u{037F}\u{1FFF}\u{200C}\u{200D}\u{2070}\u{218F}\u{2C00}\u{2FEF}\u{3001}\u{D7FF}\u{FA0E}\u{FDCF}\u{FDF0}\u{FFEF}\u{10000}\u{EFFFD}"
+        "http://a.example/?AZaz\u{E000}\u{F8FF}\u{F0000}\u{FFFFD}\u{100000}\u{10FFFD}\u{00C0}\u{00D6}\u{00D8}\u{00F6}\u{00F8}\u{02FF}\u{0370}\u{037D}\u{037F}\u{1FFF}\u{200C}\u{200D}\u{2070}\u{218F}\u{2C00}\u{2FEF}\u{3001}\u{D7FF}\u{FA0E}\u{FDCF}\u{FDF0}\u{FFEF}\u{10000}\u{EFFFD}",
+        "http://[va.12z]",
+        "http://[vff.B]",
+        "http://[V0.a]"
     ];
 
-    for e in examples.iter() {
-        let unchecked = Iri::parse_unchecked(*e);
-        match Iri::parse(*e) {
-            Ok(iri) => assert_eq!(iri, unchecked),
+    for e in examples {
+        let unchecked = Iri::parse_unchecked(e);
+        assert_eq!(unchecked, e);
+        match Iri::parse(e) {
+            Ok(iri) => {
+                assert_eq!(unchecked, iri);
+                assert_eq!(unchecked.scheme(), iri.scheme());
+                assert_eq!(unchecked.authority(), iri.authority());
+                assert_eq!(unchecked.path(), iri.path());
+                assert_eq!(unchecked.query(), iri.query());
+                assert_eq!(unchecked.fragment(), iri.fragment());
+            }
             Err(error) => panic!("{} on IRI {}", error, e),
         }
     }
@@ -135,10 +146,18 @@ fn test_relative_parsing() {
     ];
 
     let base = Iri::parse("http://a/b/c/d;p?q").unwrap();
-    for e in examples.iter() {
-        let unchecked = IriRef::parse_unchecked(*e);
-        match IriRef::parse(*e) {
-            Ok(iri) => assert_eq!(unchecked, iri),
+    for e in examples {
+        let unchecked = IriRef::parse_unchecked(e);
+        assert_eq!(unchecked, e);
+        match IriRef::parse(e) {
+            Ok(iri) => {
+                assert_eq!(unchecked, iri);
+                assert_eq!(unchecked.scheme(), iri.scheme());
+                assert_eq!(unchecked.authority(), iri.authority());
+                assert_eq!(unchecked.path(), iri.path());
+                assert_eq!(unchecked.query(), iri.query());
+                assert_eq!(unchecked.fragment(), iri.fragment());
+            }
             Err(error) => panic!("{} on relative IRI {}", error, e),
         }
         match base.resolve(e) {
@@ -234,10 +253,24 @@ fn test_wrong_relative_parsing() {
         "http://[::1]a/",
         // fuzzing bugs
         "//͏@[]",
+        "//@@",
+        "$:",
+        "-:",
+        ":",
+        // IPvFuture
+        "http://[]",
+        "http://[a]",
+        "http://[vz]",
+        "http://[v11]",
+        "http://[v1.]",
+        "http://[v.a]",
+        "http://[v1.@]",
+        "http://[v1.%01]",
+        "//[v1.ւ]",
     ];
 
     let base = Iri::parse("http://a/b/c/d;p?q").unwrap();
-    for e in examples.iter() {
+    for e in examples {
         let result = base.resolve(e);
         assert!(result.is_err(), "{} is wrongly considered valid", e);
     }
@@ -457,10 +490,10 @@ fn test_resolve_relative_iri() {
         ("y/z", "http://ex/x/y?pp/qq", "http://ex/x/y/z"),
         ("y?q", "http://ex/x/y?q", "http://ex/x/y?q"),
         ("/x/y?q", "http://ex?p", "http://ex/x/y?q"),
-        /*("c/d", "foo:a/b", "foo:a/c/d"),
+        ("c/d", "foo:a/b", "foo:a/c/d"),
         ("/c/d", "foo:a/b", "foo:/c/d"),
         ("", "foo:a/b?c#d", "foo:a/b?c"),
-        ("b/c", "foo:a", "foo:b/c"),*/
+        ("b/c", "foo:a", "foo:b/c"),
         ("../b/c", "foo:/a/y/z", "foo:/a/b/c"),
         //("./b/c", "foo:a", "foo:b/c"),
         //("/./b/c", "foo:a", "foo:/b/c"),
@@ -523,12 +556,12 @@ fn test_resolve_relative_iri() {
         ("s", "http://example.com", "http://example.com/s"),
     ];
 
-    for (relative, base, output) in examples.iter() {
-        let base = Iri::parse(*base).unwrap();
+    for (relative, base, output) in examples {
+        let base = Iri::parse(base).unwrap();
         match base.resolve(relative) {
             Ok(result) => assert_eq!(
                 result.as_str(),
-                *output,
+                output,
                 "Resolving of {relative} against {base} is wrong. Found {result} and expecting {output}"
             ),
             Err(error) => panic!(
@@ -539,12 +572,158 @@ fn test_resolve_relative_iri() {
         let result = base.resolve_unchecked(relative);
         assert_eq!(
             result.as_str(),
-            *output,
+            output,
             "Lenient resolving of {relative} against {base} is wrong. Found {result} and expecting {output}"
         );
     }
 }
 
+#[test]
+fn test_relativize_iri() {
+    let examples = [
+        ("http:", "http:", ""),
+        ("http://example.com", "http://example.com", ""),
+        ("http://example.com/foo", "http://example.com/foo", ""),
+        (
+            "http://example.com/foo/bar",
+            "http://example.com/foo/bar",
+            "",
+        ),
+        (
+            "http://example.com/foo/bar?bat",
+            "http://example.com/foo/bar?bat",
+            "",
+        ),
+        (
+            "http://example.com/foo/bar?bat#baz",
+            "http://example.com/foo/bar?bat#baz",
+            "#baz",
+        ),
+        ("http:", "https:", "http:"),
+        ("http://example.com", "http://example.org", "//example.com"),
+        ("http://example.com/foo", "http://example.com/bar", "foo"),
+        (
+            "http://example.com/foo?bat",
+            "http://example.com/foo?foo",
+            "?bat",
+        ),
+        (
+            "http://example.com/foo?bat#baz",
+            "http://example.com/foo?bat#foo",
+            "#baz",
+        ),
+        ("http://example.com", "http:", "//example.com"),
+        ("http://example.com", "http://", "//example.com"),
+        ("http://example.com", "http://example.org", "//example.com"),
+        ("http://example.com/foo", "http://example.com/", "foo"),
+        ("http://example.com/foo", "http://example.com/bar", "foo"),
+        (
+            "http://example.com/foo",
+            "http://example.com/bar/baz",
+            "/foo",
+        ),
+        (
+            "http://example.com/foo/bar",
+            "http://example.com/foo/baz",
+            "bar",
+        ),
+        (
+            "http://example.com/foo/bar",
+            "http://example.com/foo",
+            "/foo/bar",
+        ),
+        (
+            "http://example.com/foo?bar",
+            "http://example.com/foo?baz",
+            "?bar",
+        ),
+        (
+            "http://example.com?bar",
+            "http://example.com/a",
+            "//example.com?bar",
+        ),
+        ("http://example.com?bar", "http://example.com", "?bar"),
+        (
+            "http://example.com?bar",
+            "http://example.com/",
+            "//example.com?bar",
+        ),
+        (
+            "http://example.com/foo#bar",
+            "http://example.com/foo#baz",
+            "#bar",
+        ),
+        (
+            "http://example.com/foo/",
+            "http://example.com/foo/bar",
+            "/foo/",
+        ),
+        ("http://example.com/:", "http://example.com/foo", "/:"),
+        ("http:", "http://example.com", "http:"),
+        ("http:?foo", "http://example.com", "http:?foo"),
+        (
+            "http://example.com",
+            "http://example.com/foo",
+            "//example.com",
+        ),
+        (
+            "http://example.com",
+            "http://example.com?query",
+            "//example.com",
+        ),
+        (
+            "http://example.com/foo",
+            "http://example.com/foo?query",
+            "foo",
+        ),
+        ("http:?query", "http://example.com?query", "http:?query"),
+        ("http:/path", "http://example.com/foo", "http:/path"),
+        (
+            "http://example.com//a",
+            "http://example.com/",
+            "//example.com//a",
+        ),
+        ("urn:ab", "urn:", "ab"),
+        ("urn:isbn:foo", "urn:", "urn:isbn:foo"),
+        ("urn:is/bn:foo", "urn:", "is/bn:foo"),
+        ("urn:.", "urn:", "."),
+    ];
+
+    for (original, base, output) in examples {
+        let original = Iri::parse(original).unwrap();
+        let base = Iri::parse(base).unwrap();
+        let output = IriRef::parse(output).unwrap();
+        let actual = base.relativize(&original).unwrap();
+        assert_eq!(
+            actual, output,
+            "Relativizing {original} against {base} gives {actual} and not {output}"
+        );
+        let resolved = base.resolve(actual.as_str()).unwrap();
+        assert_eq!(
+            resolved, original,
+            "Resolving {actual} against {base} gives {resolved} and not {original}"
+        );
+    }
+}
+
+#[test]
+fn test_relativize_iri_fails() {
+    let examples = [
+        "http://example.com/a/../b",
+        "http://example.com/a/..",
+        "http://example.com/./b",
+        "http://example.com/.",
+    ];
+
+    let base = Iri::parse("http://example.com/s").unwrap();
+    for e in examples {
+        let e = Iri::parse(e).unwrap();
+        assert!(base.relativize(&e).is_err());
+        // We make sure it's not possible to relativize
+        assert_ne!(base.resolve(e.as_str()).unwrap(), e);
+    }
+}
+
 #[test]
 fn test_eq() {
     let iri = Iri::parse("http://example.com").unwrap();
